from sqlalchemy import create_engine, Table, MetaData
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session
from datetime import datetime
from functools import wraps
from flask import jsonify
from src.utils.logger import logger
from src.utils.exception import handle_exception
import re
from sqlalchemy import inspect

class DatabaseManager:
    
    def __init__(self, base: declarative_base, engine: create_engine):
        """
        Initialize the DatabaseHandler class.

        Args:
            base (declarative_base): The base class for the database models.
            with_session (function): The function to wrap database operations.
            db_name (str): The name of the database.
        """
        self.engine = engine
        self.base = base
        
        # First get all tables from the database
        inspector = inspect(self.engine)
        db_tables = inspector.get_table_names()
        
        # Get all models from SQLAlchemy Base
        model_tables = self.base.metadata.tables.keys()
        
        # Compare and log differences
        logger.info("=== Database Schema Validation ===")
        logger.info(f"Database tables: {sorted(db_tables)}")
        logger.info(f"Model tables: {sorted(model_tables)}")
        
        # Check for missing tables in models
        missing_in_models = set(db_tables) - set(model_tables)
        if missing_in_models:
            logger.error(f"Tables in database but missing in models: {missing_in_models}")
        
        # Check for extra tables in models
        extra_in_models = set(model_tables) - set(db_tables)
        if extra_in_models:
            logger.error(f"Tables in models but missing in database: {extra_in_models}")
            
        # Check column differences for each table
        for table_name in set(db_tables) & set(model_tables):
            db_columns = {col['name']: col for col in inspector.get_columns(table_name)}
            model_columns = {col.name: col for col in self.base.metadata.tables[table_name].columns}
            
            # Check for missing columns in models
            missing_cols = set(db_columns.keys()) - set(model_columns.keys())
            if missing_cols:
                logger.error(f"Table '{table_name}' missing columns in model: {missing_cols}")
            
            # Check for extra columns in models
            extra_cols = set(model_columns.keys()) - set(db_columns.keys())
            if extra_cols:
                logger.error(f"Table '{table_name}' has extra columns in model: {extra_cols}")
        
        try:
            self.base.metadata.create_all(self.engine)
        except Exception as e:
            logger.error(f'Error creating tables: {str(e)}')

        self.metadata = MetaData()
        self.metadata.reflect(bind=self.engine)
        logger.success(f'Database initialized')

    def with_session(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            session = Session(bind=self.engine)
            try:
                result = func(session, *args, **kwargs)
                session.commit()
                return result
            except Exception as e:
                session.rollback()
                logger.error(f"Database error in {func.__name__}: {str(e)}")
                raise Exception(f"Database error: {str(e)}")
            finally:
                session.close()
        return wrapper

    def _ids_to_string(self, data: dict):
        for key, value in data.items():
            if re.match(r'^id$|^\w+_id$', key):
                data[key] = str(value)
        return data
    
    def _none_to_null(self, data: dict):
        for key, value in data.items():
            if value == 'None':
                data[key] = None
        return data

    def _dates_to_timestamp(self, data: dict):
        for key, value in data.items():
            if isinstance(value, datetime):
                data[key] = value.strftime('%Y%m%d%H%M%S')
            elif isinstance(value, str):
                try:
                    data[key] = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%fZ').strftime('%Y%m%d%H%M%S')
                except:
                    pass
        return data

    def create(self, table: str, data: dict = None) -> str:
        @self.with_session
        def _create(session, table: str, data: dict = None):
            logger.info(f'Attempting to create new entry in table: {table}')

            if not data:
                raise Exception("Data to create must be provided.")
            
            # Get the model class for the table
            model = None
            for class_ in self.base.__subclasses__():
                if hasattr(class_, '__tablename__') and class_.__tablename__ == table:
                    model = class_
                    break
            
            if not model:
                raise Exception(f"Model not found for table: {table}")
            
            current_time = datetime.now().strftime('%Y%m%d%H%M%S')
            data = self._dates_to_timestamp(data)

            data = {
                'created': current_time,
                'updated': current_time,
                **data
            }
            
            # Create a new instance of the model
            new_record = model(**data)
            session.add(new_record)
            session.flush()
            
            logger.success(f'Successfully created entry with id: {new_record.id}')
            return str(new_record.id)

        return _create(table, data)

    def read(self, table: str, query: dict = None) -> list:
        @self.with_session
        def _read(session, table: str, query: dict = None):

            logger.info(f'Attempting to read entry from table: {table} with query: {query}')

            if query is None:
                raise Exception("Query must be provided.")
            
            tbl = Table(table, self.metadata, autoload_with=self.engine)
            sql_query = session.query(tbl)

            if query:
                for key, value in query.items():
                    if hasattr(tbl.c, key):
                        sql_query = sql_query.filter(getattr(tbl.c, key) == value)
                
            # Print the generated SQL query
            compiled_query = sql_query.statement.compile(
                compile_kwargs={"literal_binds": True},
                dialect=self.engine.dialect
            )
            logger.info(f'Generated SQL: {str(compiled_query)}')
            
            results = sql_query.all()

            serialized_results = [row._asdict() for row in results]

            for result in serialized_results:
                self._ids_to_string(result)
                self._none_to_null(result)
            logger.success(f'Successfully read {len(serialized_results)} entries from table: {table}')
            return serialized_results

        return _read(table, query)

    def update(self, table: str, query: dict = None, data: dict = None) -> str:
        @self.with_session
        def _update(session, table: str, query: dict = None, data: dict = None):
            logger.info(f'Attempting to update entry in table: {table}')

            if query is None:
                raise Exception("Query must be provided.")
            
            if data is None:
                raise Exception("Data to update must be provided.")

            tbl = Table(table, self.metadata, autoload_with=self.engine)
            sql_query = session.query(tbl)

            for key, value in query.items():
                if hasattr(tbl.c, key):
                    sql_query = sql_query.filter(getattr(tbl.c, key) == value)

            item = sql_query.first()

            if not item:
                raise Exception(f"{table.capitalize()} with given parameters not found")
            
            logger.info(f'Updating entry timestamp.')
            data['updated'] = datetime.now().strftime('%Y%m%d%H%M%S')
            data = self._dates_to_timestamp(data)

            sql_query.update(data)
            session.flush()

            updated_item = sql_query.first()
            logger.success(f"Successfully updated entry with id: {updated_item.id} in table: {table}.")
            
            return str(updated_item.id)

        return _update(table, query, data)

    def delete(self, table: str, query: dict = None) -> str:
        @self.with_session
        def _delete(session, table: str, query: dict = None):
            logger.info(f'Attempting to delete entry from table: {table}')

            if query is None:
                raise Exception("Query must be provided.")
            
            tbl = Table(table, self.metadata, autoload_with=self.engine)
            sql_query = session.query(tbl)

            if query:
                for key, value in query.items():
                    if hasattr(tbl.c, key):
                        sql_query = sql_query.filter(getattr(tbl.c, key) == value)

            item = sql_query.first()
            if not item:
                raise Exception(f"Entry with given parameters not found in table: {table}.")

            delete_stmt = tbl.delete().where(tbl.c.id == item.id)
            session.execute(delete_stmt)
            session.flush()

            logger.success(f"Successfully deleted entry with id: {item.id} from table: {table}.")
            return str(item.id)

        return _delete(table, query)
        
    def get_tables(self):
        @self.with_session
        def _get_tables(session):
            """Returns a list of all tables in the database."""
            logger.info('Attempting to get all tables from database')
            table_names = self.metadata.tables.keys()
            logger.success(f'Successfully retrieved {len(table_names)} tables')
            return list(table_names)

        return _get_tables()

    @handle_exception
    def get_schema(self, table: str):
        """Returns the schema of a specified table."""
        logger.info(f'Attempting to get schema for table: {table}')
        if table not in self.metadata.tables:
            raise Exception(f"Table '{table}' not found in database")
        
        tbl = self.metadata.tables[table]
        schema = {}
        
        for column in tbl.columns:
            schema[column.name] = {
                'type': str(column.type),
                'nullable': column.nullable,
                'primary_key': column.primary_key,
                'default': str(column.default) if column.default else None,
                'foreign_keys': [str(fk.target_fullname) for fk in column.foreign_keys]
            }
        
        logger.success(f'Successfully retrieved schema for table: {table}')
        return schema

    def from_data_object(self, data: dict, table: str, overwrite: bool = False):
        """
        Imports a data object to a SQLite table.
        The data object must be a list of dictionaries [{}, {}, {}]. (pd.DataFrame.to_dict('records') format)
        Recieves data and destination table name, and imports the data to the table.
        If overwrite is True, the table will be truncated before the data is imported.
        If the table does not exist, it will be created.
        """
        @self.with_session
        @handle_exception
        def _from_data_object(session, data: dict, table: str, overwrite: bool):
            logger.info(f'Attempting to import data to table: {table}')
            try:
                if not isinstance(data, list) or not all(isinstance(item, dict) for item in data):
                    raise Exception("Data must be a list of dictionaries")

                tbl = Table(table, self.metadata, autoload_with=self.engine)

                if overwrite:
                    logger.info(f'Truncating table: {table}')
                    session.execute(tbl.delete())

                if not data:
                    logger.warning(f'No data to import to table: {table}')
                    return jsonify({'inserted': 0})

                current_time = datetime.now().strftime('%Y%m%d%H%M%S')
                for item in data:
                    item['created'] = current_time
                    item['updated'] = current_time

                session.execute(tbl.insert(), data)
                session.flush()
                
                count = len(data)
                logger.success(f'Successfully imported {count} records to table: {table}')
                return jsonify({'inserted': count})

            except SQLAlchemyError as e:
                logger.error(f'Error importing data: {str(e)}')
                raise Exception(f'Database error: {str(e)}')

        return _from_data_object(data, table, overwrite)


# laserfocus/__init__.py
from .utils import exception, logger
from .utils.managers.database import DatabaseManager
from .utils.managers.secrets import SecretsManager
from .utils.managers.scopes import ScopesManager


import pytest
from unittest.mock import patch, MagicMock
from laserfocus.utils.logger import Logger

@pytest.fixture
def logger():
    return Logger()

def test_logger_initialization(logger):
    assert logger.console is not None
    assert logger.logger is not None

@patch('logging.Logger.debug')
def test_logger_info(mock_debug, logger):
    test_message = "test info message"
    logger.info(test_message)
    mock_debug.assert_called_once_with(
        f"[blue]{test_message}[/blue]",
        extra={'markup': True}
    )

@patch('logging.Logger.debug')
def test_logger_success(mock_debug, logger):
    test_message = "test success message"
    logger.success(test_message)
    mock_debug.assert_called_once_with(
        f"[green]{test_message}[/green]",
        extra={'markup': True}
    )

@patch('logging.Logger.info')
def test_logger_announcement_info(mock_info, logger):
    test_message = "test announcement"
    logger.announcement(test_message, type='info')
    mock_info.assert_called_once_with(
        f"[red bold]{test_message}[/red bold]",
        extra={'markup': True}
    )

@patch('logging.Logger.info')
def test_logger_announcement_success(mock_info, logger):
    test_message = "test announcement"
    logger.announcement(test_message, type='success')
    mock_info.assert_called_once_with(
        f"[white bold]{test_message}[/white bold]\n",
        extra={'markup': True}
    )

def test_logger_announcement_invalid_type(logger):
    with pytest.raises(ValueError, match="Invalid type. Choose 'info' or 'success'."):
        logger.announcement("test", type='invalid')

@patch('logging.Logger.warning')
def test_logger_warning(mock_warning, logger):
    test_message = "test warning message"
    logger.warning(test_message)
    mock_warning.assert_called_once_with(
        f"[yellow bold on white]{test_message}[/yellow bold on white]",
        extra={'markup': True}
    )

@patch('logging.Logger.error')
def test_logger_error(mock_error, logger):
    test_message = "test error message"
    logger.error(test_message)
    mock_error.assert_called_once_with(
        f"[red bold on white]{test_message}[/red bold on white]",
        extra={'markup': True}
    ) 


from .logger import logger
from flask import Response
import json
import functools

def handle_exception(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            data = func(*args, **kwargs)
            return data
        except Exception as e:
            logger.error(f"Error in {func.__name__}: {e}")
            json_data = json.dumps({'error': str(e)})
            return Response(json_data, status=500, mimetype='application/json')
    return wrapper


import pytest
from laserfocus.utils.response import Response

def test_response_init():
    response = Response("success", "test content")
    assert response.status == "success"
    assert response.content == "test content"

def test_response_success():
    response_dict = Response.success("test success")
    assert response_dict["status"] == "success"
    assert response_dict["content"] == "test success"

def test_response_error():
    response_dict = Response.error("test error")
    assert response_dict["status"] == "error"
    assert response_dict["content"] == "test error"

def test_response_to_dict():
    response = Response("custom", {"key": "value"})
    response_dict = response.to_dict()
    assert response_dict == {
        "status": "custom",
        "content": {"key": "value"}
    } 


import pytest
from sqlalchemy import create_engine, Column, Integer, String, DateTime
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session
from datetime import datetime
from laserfocus.utils.database import DatabaseHandler
from laserfocus.utils.response import Response

# Test setup
Base = declarative_base()

class TestModel(Base):
    __tablename__ = 'test_table'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    created = Column(DateTime)
    updated = Column(DateTime)

@pytest.fixture
def db_handler():
    engine = create_engine('sqlite:///:memory:')
    handler = DatabaseHandler(Base, engine, type='sqlite')
    return handler

def test_database_initialization(db_handler):
    assert db_handler.engine is not None
    assert db_handler.type == 'sqlite'
    assert db_handler.base is not None
    assert db_handler.metadata is not None

def test_create_record(db_handler):
    data = {'name': 'test_name'}
    response = db_handler.create('test_table', data)
    
    assert response['status'] == 'success'
    assert isinstance(response['content'], int)  # Should return the new ID

    # Verify the record exists
    read_response = db_handler.read('test_table', {'id': response['content']})
    assert read_response['status'] == 'success'
    assert len(read_response['content']) == 1
    assert read_response['content'][0]['name'] == 'test_name'

def test_update_record(db_handler):
    # First create a record
    data = {'name': 'original_name'}
    create_response = db_handler.create('test_table', data)
    record_id = create_response['content']

    # Update the record
    update_data = {'name': 'updated_name'}
    update_response = db_handler.update('test_table', {'id': record_id}, update_data)
    
    assert update_response['status'] == 'success'
    
    # Verify the update
    read_response = db_handler.read('test_table', {'id': record_id})
    assert read_response['content'][0]['name'] == 'updated_name'

def test_read_records(db_handler):
    # Create multiple records
    db_handler.create('test_table', {'name': 'test1'})
    db_handler.create('test_table', {'name': 'test2'})
    
    # Read all records
    response = db_handler.read('test_table')
    assert response['status'] == 'success'
    assert len(response['content']) == 2
    
    # Read with filter
    filtered_response = db_handler.read('test_table', {'name': 'test1'})
    assert filtered_response['status'] == 'success'
    assert len(filtered_response['content']) == 1
    assert filtered_response['content'][0]['name'] == 'test1'

def test_delete_record(db_handler):
    # Create a record
    data = {'name': 'to_delete'}
    create_response = db_handler.create('test_table', data)
    record_id = create_response['content']
    
    # Delete the record
    delete_response = db_handler.delete('test_table', {'id': record_id})
    assert delete_response['status'] == 'success'
    
    # Verify deletion
    read_response = db_handler.read('test_table', {'id': record_id})
    assert len(read_response['content']) == 0

def test_delete_all_records(db_handler):
    # Create multiple records
    db_handler.create('test_table', {'name': 'test1'})
    db_handler.create('test_table', {'name': 'test2'})
    
    # Delete all records
    response = db_handler.delete_all('test_table')
    assert response['status'] == 'success'
    
    # Verify all records are deleted
    read_response = db_handler.read('test_table')
    assert len(read_response['content']) == 0

def test_get_tables(db_handler):
    response = db_handler.get_tables()
    assert response['status'] == 'success'
    assert 'test_table' in response['content']

def test_get_schema(db_handler):
    response = db_handler.get_schema('test_table')
    assert response['status'] == 'success'
    schema = response['content']
    
    assert 'id' in schema
    assert 'name' in schema
    assert 'created' in schema
    assert 'updated' in schema
    
    assert schema['id']['primary_key'] is True
    assert schema['name']['type'] == 'VARCHAR'

def test_from_data_object(db_handler):
    data = [
        {'name': 'bulk1'},
        {'name': 'bulk2'}
    ]
    
    response = db_handler.from_data_object(data, 'test_table', overwrite=True)
    assert response['status'] == 'success'
    
    # Verify data was imported
    read_response = db_handler.read('test_table')
    assert len(read_response['content']) == 2
    names = [record['name'] for record in read_response['content']]
    assert 'bulk1' in names
    assert 'bulk2' in names

def test_error_handling(db_handler):
    # Test invalid table name
    response = db_handler.read('nonexistent_table')
    assert response['status'] == 'error'
    assert 'Database error' in response['content']
    
    # Test invalid column name in filter
    response = db_handler.read('test_table', {'nonexistent_column': 'value'})
    assert response['status'] == 'success'  # Should ignore invalid column
    
    # Test invalid data type
    response = db_handler.from_data_object("not a list", 'test_table')
    assert response['status'] == 'error'
    assert 'Data must be a list of dictionaries' in response['content'] 


from functools import wraps
from flask import request, jsonify
from flask_jwt_extended import get_jwt, get_jwt_identity
from logger import logger

public_routes = ['docs', 'index', 'token', 'oauth.login', 'oauth.create']

def verify_scope(required_scope):
    def decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            logger.info(f'Verifying if user has {required_scope} scope')
            if request.endpoint in public_routes:
                return fn(*args, **kwargs)
            
            claims = get_jwt()
            user_scopes = claims.get('scopes', '').split()
            
            # "all" scope has access to everything
            if 'all' in user_scopes:
                logger.success(f'User has all scope, granting access')
                return fn(*args, **kwargs)
            
            # Check if user has the exact scope or a parent scope
            scope_parts = required_scope.split('/')
            for i in range(len(scope_parts)):
                partial_scope = '/'.join(scope_parts[:i+1])
                if partial_scope in user_scopes:
                    logger.success(f'User has {partial_scope} scope, granting access')
                    return fn(*args, **kwargs)
            
            logger.warning(f'User attempted to access {required_scope} without proper authorization. User scopes: {user_scopes}')
            return jsonify({"error": "Insufficient scope"}), 403
        return wrapper
    return decorator

def enforce_user_filter(field_name='user_id'):
    """
    Decorator that enforces user-specific filtering for routes.
    This decorator should be applied after @verify_scope.
    
    :param field_name: The field name to filter by (defaults to 'user_id')
    """
    def decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            claims = get_jwt()
            current_user = get_jwt_identity()
            user_scopes = claims.get('scopes', '').split()
            logger.info(f'Enforcing user filter for current user: {current_user}')
            
            # If user has 'all' scope, no filtering needed
            if 'all' in user_scopes:
                logger.success(f'User has all scope, no filtering needed')
                return fn(*args, **kwargs)
            
            # Get the request payload
            payload = request.get_json(force=True)
            
            # Handle different HTTP methods and payload structures
            if request.method == 'POST':
                # For create operations, add user ID to data
                if 'data' in payload:
                    payload['data'][field_name] = current_user
                
                # For read/update operations with query
                if 'query' in payload:
                    if not isinstance(payload['query'], dict):
                        payload['query'] = {}
                    payload['query'][field_name] = current_user
                else:
                    # If no query exists, create one with the user filter
                    payload['query'] = {field_name: current_user}
            
            logger.success(f'Enforced user filter for {field_name}')
            return fn(*args, **kwargs)
            
        return wrapper
    return decorator
       


from rich.logging import RichHandler
from rich.console import Console
from rich.theme import Theme
import logging

class Logger:
    def __init__(self):
        
        custom_theme = Theme({
            "primary": "#FA3232",
            "secondary": "#0EA5E9",
            "warning": "#FFB300",
            "success": "#00C853",
            "error": "#FF4B4B",
            "white": "#FFFFFF",
            "bold primary": "#FA3232",
            "bold secondary": "#0EA5E9",
            "bold warning": "#FFB300",
            "bold success": "#00C853",
            "bold error": "#FF4B4B",
            "bold white": "#FFFFFF",
        })

        self.console = Console(
            theme=custom_theme,
            color_system="truecolor"
        )
        
        logging.basicConfig(
            level=logging.DEBUG,
            format="%(message)s",
            datefmt="[%X]",
            handlers=[RichHandler(console=self.console, rich_tracebacks=True)]
        )
        self.logger = logging.getLogger("rich")

        # Suppress other logs
        logging.getLogger('werkzeug').setLevel(logging.ERROR)
        logging.getLogger('flask').setLevel(logging.ERROR)
        logging.getLogger('requests').setLevel(logging.ERROR)
        logging.getLogger('connectionpool').setLevel(logging.ERROR)
        logging.getLogger('urllib3.connectionpool').setLevel(logging.ERROR)
        logging.getLogger('googleapiclient.discovery').setLevel(logging.ERROR)
        logging.getLogger('google_auth_httplib2').setLevel(logging.ERROR)
        logging.getLogger('feedparser').setLevel(logging.ERROR)
        logging.getLogger('nltk').setLevel(logging.ERROR)
        logging.getLogger('chardet').setLevel(logging.ERROR)
        logging.getLogger('sqlalchemy').setLevel(logging.ERROR)
        logging.getLogger('ib_insync').setLevel(logging.ERROR)
        logging.getLogger('selector_events').setLevel(logging.ERROR)
        logging.getLogger('google-auth').setLevel(logging.ERROR)
        logging.getLogger('google.auth.transport.requests').setLevel(logging.ERROR)
        logging.getLogger('geventwebsocket').setLevel(logging.ERROR)

    def info(self, message):
        self.logger.debug(f"[white]{message}[/white]", extra={'markup': True})

    def success(self, message):
        self.logger.debug(f"[success]{message}[/success]", extra={'markup': True})

    def announcement(self, message, type='info'):
        if type == 'info':
            self.logger.info(f"[bold secondary]{message}[/bold secondary]", extra={'markup': True})
        elif type == 'success':
            self.logger.info(f"[bold primary]{message}[/bold primary]\n", extra={'markup': True})
        else:
            raise ValueError("Invalid type. Choose 'info' or 'success'.")
        
    def warning(self, message):
        self.logger.warning(f"[bold warning]{message}[/bold warning]", extra={'markup': True})

    def error(self, message):
        self.logger.error(f"[on white][error]{message}[/error][/on white]", extra={'markup': True})


logger = Logger()



# setup.py
from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="laserfocus",  # Replace with your package name
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    author="laserfocus",
    author_email="aa@laserfocus.space",
    description="A collection of utilities for laserfocus",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/laserfocus/pip",  # Optional: GitHub URL
    packages=find_packages(),  # Automatically finds your package
    install_requires=[
        "Requests==2.32.3",
        "certifi==2024.7.4",
        "charset-normalizer==3.4.1",
        "idna==3.10",
        "markdown-it-py==3.0.0",
        "mdurl==0.1.2",
        "Pygments==2.19.1",
        "urllib3==2.3.0",
        "rich==13.9.3",
        "SQLAlchemy==2.0.35",
        "beautifulsoup4==4.12.3",
        "soupsieve==2.6",
        "typing_extensions==4.12.2"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",  # Specify compatible Python versions
)